/*
** This program is distributed with the FALKEN BBS and with the FALKEN
** Doors Development Kit.  This program is copyright 1989 by Herb Rose
**
** An attempt has been made to document this program to the extent that
** a person familiar with C can comprehend the steps necessary to write
** a doors program for FALKEN
*/

/*
** First, the general include files and some data declarations
*/

#include "stdio.h"
#include "structs.h"

int endtask();

char menu2[]=
"<ENTER> to continue, 'S' to stop, 'C' for continuous output>";

char active[]="FALKEN Bulletins Section.\r";

/*
** By setting up a typedef to represent an array of account structures
** and user structures, we can easily gain access to individual
** account information by declaring a pointer to these data types later.
*/

typedef struct acct_rec acctp[];
typedef struct user_rec userp[];

char workbuf[5120];

char menutext[4096];

/*
** This template is used to work with the initialization message
** sent by FALKEN when the program is started.
**
** Note that we set these templates up and declare a pointer to the
** structure.  The character array WORKBUF, above, is where all
** incoming messages will be placed.  By loading the strucure
** pointers with the address of the first byte of WORKBUF, we can
** use these templates to access the information in the correct
** formats.
*/

struct msg3 {   /* the initial message with line num and queue nums */
    int m3typ;   /* must be 7 */
    int m3line;
    int m3inq;
    int m3outq;
    acctp *acctptr;
    userp *userptr;
    int m3maxline;
    } *m3;

/*
** This template is used to send text back and forth between us and the
** FALKEN task.
*/

struct msg1 {  /* used for text in/out */
    int  m1type;
    int  m1num;  /* line number */
    char m1text[5000];
    } *m1;

char cb[100];

int who;
int inq, outq;

int *iptr;

int nostop=0;

main()
{
    int j,k,n,len;
    int k2;

    FILE *fp;

    acctp *acct;    /* pointer to an array of account structures */
    userp *user;    /* pointer to an array of user_rec structures */

    char path[60];

    iptr= (int *)workbuf;

    memset(path,0,60);

    m1=(struct msg1 *)workbuf;
    m3=(struct msg3 *)workbuf;

/*
** wait for the initialization event from FALKEN.
** it will be event type 7.
** If for some reason we get an event type 8, we must terminate.
*/
    m3->m3typ=0;
    while(m3->m3typ != 7)
    {
        recvmsg(1,workbuf,1000);
        if(m3->m3typ == 8) endtask();
    }

/*
** load up our structure pointers with the address of the structure
** arrays in memory
*/

    acct=m3->acctptr;
    user=m3->userptr;
/*
** get the input and output queue numbers we have been assigned,
** and which line this user is logged on to.
*/

    inq=m3->m3inq;
    outq=m3->m3outq;
    who=m3->m3line;

/*
** send() is a local function used to transmit a null-terminated
** string to a user.  We have to specify the user we wish to send
** the string to, the string address, and which queue we are using
** to get the message to FALKEN.
** Only the FALKEN program can actually send data to the user, so this
** message is actually telling FALKEN to send this data to the user
** for us.
*/

    send(who,active,outq);  /* send a 'we are alive' to the user */

/*
**  Now the main loop -
**   basically, we wait for the user to type something.
**   anything except 'exit' is considered to be a file name.
**   if 'exit' is entered, we will terminate normally.
**   else, attempt to open the file, and send it to the user.
**   we will pause between pages.
*/

    if((*user)[who].u_ansi)
    {
        fp=fopen("bullmenu.ans","rb");
    }
    else
    {
        fp=fopen("bullmenu.txt","rb");
    }
    memset(menutext,0,4096);
    fread(menutext,1,4096,fp);
    fclose(fp);
/*
** get rid of the LF's in the menu text
*/
    for(j=k=0; menutext[j]; j++)
    {
        if(menutext[j]!='\n') menutext[k++]=menutext[j];
    }
    menutext[k]='\0';   /* make sure we get the last null */

    for(;;) {
        send(who,menutext,outq);
/*
** recvmsg extracts the first message from the specified queue.
** note that we receive the message in WORKBUF.  Since our
** structure template pointers are pointing to workbuf, we can
** easily check for message type and access the portions of the
** message we need without copying things around.
*/

        recvmsg(inq,workbuf,1024);
/*
** always watchful for the 'terminate now' command
*/
        if(m1->m1type == 8) endtask();
/*
** if message type is 1 (normal text received),
** and the byte count is > 0
** then assume it is either the 'exit' command or a file name
*/
        else if((m1->m1type==1)&&(m1->m1num)) {
            strlwr(m1->m1text);
            if((m1->m1text[0]=='q')||(m1->m1text[0]=='x')) endtask();
            j=atoi(m1->m1text);
            if(j<1 || j>16)
            {
                send(who,
                   "Invalid Bulletin Number.  Press [ENTER] to continue.\r",
                    outq);
                recvmsg(inq,workbuf,1024);
                continue;
            }
            (*acct)[who].bulreadflags&=~(1<<(j-1));
            sprintf(cb,"bulletin.%d",j);
            fp=fopen(cb,"ra");
            if(fp==NULL) {      /* cannot open file! */
                send(who,"Error opening file.\r",outq);
                }
            else {  /* file is open, so we can start sending it to user */
                nostop=0;
                n=0;
/*
** This while loop causes us to loop until end of file, or until
** the user enters 'stop'.  At that point we go back to the
** outer loop, that continues until the user enters 'exit' to
** terminate the program
*/
                while(!n) {
                    for(k=0; k<23;) {
                        if(fgets(cb,79,fp)==NULL) {
                            n=1;
                            send(who,
                              "End of bulletin.  Press [ENTER] to continue\r",
                               outq);
                            recvmsg(inq,workbuf,1024);
                            break;
                            }
                        else {
                            cb[79]='\0';
                            len=strlen(cb);
                            cb[len-1]='\r';
/*
** the screen width of the user's terminal is in the account structure
** as member 'linlen'.  We can determine how many screen lines this
** string will create on the user's screen by dividing the length by
** his screen width.  Add one for the \r at the end
** we will continue sending output until we have sent 23 screen lines.
*/
                            if(len>(*acct)[who].linlen)k++;
                            k++;
                            send(who,cb,outq);
                            }
                        } /* end for */
                    if(!n) {
                        if(nostop) {
                            k=waitforempty();
                            if(k) break;    /* break the while loop */
                            }
                        else {
                            send(who,menu2,outq);
                            recvmsg(inq,workbuf,1024);
                            if((m1->m1type==1)&&(m1->m1num)) {
                                strlwr(m1->m1text);
                                if(m1->m1text[0]=='s') n=1;
                                else if(m1->m1text[0]=='c')
                                    nostop=1;
                                }
                            else if(m1->m1type == 8) {
                                fclose(fp);
                                endtask();
                                }
                            } /* end else nostop */
                        } /* end if !n */
                    } /* end while */
                fclose(fp);
                } /* end else */
            } /* end if */
        } /* end for (;;) */
}

waitforempty()
{
    int j,k;
    for(;;) {
        m1->m1type=12;   /* get output buffer available count */
        send_md_msg(outq,workbuf,4);
        recvmsg(inq,workbuf,1000);
        switch(m1->m1type) {
            case 8 :
                endtask();
                break;
            case 6 :
                if(m1->m1num > 2500) return 0;
                break;
            case 1 :
                if((m1->m1num) && (strcmpi("stop",m1->m1text)==0)) return 1;
                break;
            }
        }
}

/*
** Utility routines follow
**
** These routines call Multidos interface routines which are available
** with the FALKEN  Doors development kit.
**
** See the file MDOS.TXT for a description of the interface routines.
*/

/*
** Send a null terminated string to FALKEN, to be transmitted to
** a particular line;
*/

send(who,str,oq) int who; char *str; int oq;
{
    strcpy(m1->m1text,str);
    m1->m1num=who;
    m1->m1type=1;
    send_md_msg(oq,workbuf,strlen(str)+5);
/* 5 = sizeof(mltype) + sizeof(mlnum) + terminating null */
}

/*
** wait for a message on the input queue named.
** if no message, relinquish our time slice.
** after all, no need to just loop here when other tasks
** can use the processor.
*/


/*
** This program gracefully exits by
** sending the 'task has terminated' message to FALKEN,
** then terminating.
*/
int endtask()
{
    *iptr=4;
    send_md_msg(outq,workbuf,2);
    exit(0);
}
    
